@@ -1,13 +1,15 @@
 from .normalization import (
     ens_process,
     ens_normalize,
-    ens_force_normalize,
+    ens_cure,
     ens_beautify,
     ens_tokenize,
     ens_transformations,
     is_ens_normalized,
-    DisallowedLabelError,
-    DisallowedLabelErrorType,
+    DisallowedNameError,
+    DisallowedNameErrorType,
+    CurableError,
+    CurableErrorType,
     NormalizationTransformation,
     NormalizationTransformationType,
     TY_VALID,
